# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Collector Widget.
"""
from __future__ import print_function

import logging, os
import numpy as np
import numpy.ma as ma

from argos.collect.collectortree import CollectorTree, CollectorSpinBox, SpinSlider
from argos.inspector.abstract import UpdateReason
from argos.qt import Qt, QtWidgets, QtGui, QtCore, QtSignal, QtSlot
from argos.repo.baserti import BaseRti
from argos.utils.cls import check_class, check_is_a_sequence, check_is_an_array
from argos.utils.masks import ArrayWithMask
from argos.utils.moduleinfo import versionStrToTuple
from argos.widgets.constants import TOP_DOCK_HEIGHT, DOCK_SPACING, DOCK_MARGIN
from argos.widgets.misc import BasePanel


logger = logging.getLogger(__name__)

USE_SLIDER = True  # SpinSliders or only Spinboxes

FAKE_DIM_NAME = '-'     # The name of the fake dimension with length 1
FAKE_DIM_OFFSET = 1000  # Fake dimensions start here (so all arrays must have a smaller ndim)


# Qt classes have many ancestors
#pylint: disable=R0901



class Collector(BasePanel):
    """ Widget for collecting the selected data.
        Consists of a table to collect the RtiItems

        The CollectorTree only stores the items, the intelligence is located in the Collector
        itself.
    """
    sigContentsChanged = QtSignal(str) # one of the UpdateReason values.
    sigShowMessage = QtSignal(str)

    def __init__(self, windowNumber):
        """ Constructor
        """
        super(Collector, self).__init__()

        self._windowNumber = windowNumber
        self._rti = None
        self._rtiInfo = None

        self._signalsBlocked = False
        self.COL_FIRST_COMBO = 1     # Column that contains the first (left most) combobox
        self.AXIS_POST_FIX = "-axis" # Added to the axis label to give the combobox labels.
        self._axisNames = []         # Axis names. Correspond to the independent variables
        self._fullAxisNames = []     # Will be set in clearAndSetComboBoxes
        self._comboBoxes = []        # Will be set in clearAndSetComboBoxes
        self._spinBoxes = []         # Will be set in createSpinBoxes
        self._spinSliders = []       # Will be set in createSpinBoxes

        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setSpacing(DOCK_SPACING)
        self.layout.setContentsMargins(DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN)
        self.buttonLayout = QtWidgets.QVBoxLayout()
        # self.layout.setContentsMargins(2, 0, 2, 0)

        # Add tree
        self.tree = CollectorTree(self)
        self.layout.addWidget(self.tree)

        # Add buttons (not yet implemented)
        # self.addVisItemButton = QtWidgets.QPushButton("Add")
        # self.addVisItemButton.setEnabled(False) # not yet implemented
        # self.buttonLayout.addWidget(self.addVisItemButton, stretch=0)
        # self.removeVisItemButton = QtWidgets.QPushButton("Remove")
        # self.removeVisItemButton.setEnabled(False) # not yet implemented
        # self.buttonLayout.addWidget(self.removeVisItemButton, stretch=0)
        # self.buttonLayout.addStretch(stretch=1)
        # self.layout.addLayout(self.buttonLayout, stretch=0)

        self._updateRtiInfo()


    def sizeHint(self):
        """ The recommended size for the widget."""
        return QtCore.QSize(300, TOP_DOCK_HEIGHT)

    @property
    def windowNumber(self):
        """ The instance number of the window this collector belongs to.
        """
        return self._windowNumber

    @property
    def rti(self):
        """ The current repository tree item. Can be None.
        """
        return self._rti


    @property
    def rtiIsSliceable(self):
        """ Returns true if the RTI is not None and is sliceable
        """
        return self._rti and self._rti.isSliceable


    @property
    def axisNames(self):
        """ Returns a the axis names.
            The axis names indicate the independent dimensions of the visualization.
        """
        return self._axisNames


    @property
    def fullAxisNames(self):
        """ Returns the of the combobox labels.
            The combobox labes the axisNames followed by '-axis'
        """
        return self._fullAxisNames


    @property
    def maxCombos(self):
        """ Returns the maximum number of combo boxes """
        return len(self._fullAxisNames)


    # @property
    # def numCombos(self):
    #     """ Returns the number of combo boxes """
    #     return len(self._comboBoxes)


    def blockChildrenSignals(self, block):
        """ If block equals True, the signals of the combo boxes and spin boxes are blocked
            Returns the old blocking state.
        """
        logger.debug("Blocking collector signals")
        for spinBox in self._spinBoxes:
            spinBox.blockSignals(block)
        for comboBox in self._comboBoxes:
            comboBox.blockSignals(block)
        result = self._signalsBlocked
        self._signalsBlocked = block
        return result


    def _setColumnCountForContents(self):
        """ Sets the column count given the current axes and selected RTI.
            Returns the newly set column count.
        """
        numRtiDims = self.rti.nDims if self.rti and self.rti.isSliceable else 0

        colCount = self.COL_FIRST_COMBO + max(numRtiDims, len(self.axisNames))
        self.tree.model().setColumnCount(colCount)
        return colCount


    def clear(self):
        """ Removes all widgets
        """
        model = self.tree.model()
        # Don't use model.clear(). it will delete the column sizes
        model.removeRows(0, 1)
        model.setRowCount(1)
        self._setColumnCountForContents()


    def clearAndSetComboBoxes(self, axesNames):
        """ Removes all comboboxes.
        """
        logger.debug("Collector clearAndSetComboBoxes: {}".format(axesNames))
        check_is_a_sequence(axesNames)
        row = 0
        self._deleteComboBoxes(row)
        self.clear()
        self._setAxesNames(axesNames)
        self._createComboBoxes(row)
        self._updateWidgets()


    def _setAxesNames(self, axisNames):
        """ Sets the axesnames, combobox lables and updates the headers. Removes old values first.
            The comboLables is the axes name + '-axis'
        """
        for col, _ in enumerate(self._fullAxisNames, self.COL_FIRST_COMBO):
            self._setHeaderLabel(col, '')

        self._axisNames = tuple(axisNames)
        self._fullAxisNames = tuple([axName + self.AXIS_POST_FIX for axName in axisNames])

        for col, label in enumerate(self._fullAxisNames, self.COL_FIRST_COMBO):
            self._setHeaderLabel(col, label)


    def _setHeaderLabel(self, col, text):
        """ Sets the header of column col to text.
            Will increase the number of columns if col is larger than the current number.
        """
        model = self.tree.model()
        item = model.horizontalHeaderItem(col)
        if item:
            item.setText(text)
        else:
            model.setHorizontalHeaderItem(col, QtGui.QStandardItem(text))


    @QtSlot(BaseRti)
    def setRti(self, rti):
        """ Updates the current VisItem from the contents of the repo tree item.

            Is a slot but the signal is usually connected to the Collector, which then calls
            this function directly.
        """
        check_class(rti, BaseRti)
        #assert rti.isSliceable, "RTI must be sliceable" # TODO: maybe later

        self._rti = rti
        self._updateWidgets()
        self._updateRtiInfo()


    def _updateWidgets(self):
        """ Updates the combo and spin boxes given the new rti or axes.
            Emits the sigContentsChanged signal.
        """
        row = 0
        model = self.tree.model()

        # Create path label
        nodePath = '' if self.rti is None else self.rti.nodePath
        pathItem = model.item(row, 0)
        if pathItem:
            # Reusing path item. If we remove all items the column size will be lost.
            pathItem.setText(nodePath)
        else:
            pathItem = QtGui.QStandardItem(nodePath)
            pathItem.setEditable(False)
            model.setItem(row, 0, pathItem)

        pathItem.setToolTip(nodePath)
        if self.rti is not None:
            pathItem.setIcon(self.rti.decoration)

        self._deleteSpinBoxes(row)
        self._populateComboBoxes(row)
        self._createSpinBoxes(row)
        self._updateRtiInfo()

        self.tree.resizeColumnsFromContents(startCol=self.COL_FIRST_COMBO)

        logger.debug("{} sigContentsChanged signal (_updateWidgets)"
                      .format("Blocked" if self.signalsBlocked() else "Emitting"))
        self.sigContentsChanged.emit(UpdateReason.RTI_CHANGED)


    def _createComboBoxes(self, row):
        """ Creates a combo box for each of the fullAxisNames
        """
        tree = self.tree
        model = self.tree.model()
        self._setColumnCountForContents()

        for col, _ in enumerate(self._axisNames, self.COL_FIRST_COMBO):
            logger.debug("Adding combobox at ({}, {})".format(row, col))
            comboBox = QtWidgets.QComboBox()
            comboBox.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
            comboBox.setMinimumContentsLength(3) # So that collector table header always readable
            comboBox.activated.connect(self._comboBoxActivated)
            self._comboBoxes.append(comboBox)

            #editor = LabeledWidget(QtWidgets.QLabel(comboLabel), comboBox)
            tree.setIndexWidget(model.index(row, col), comboBox)


    def _deleteComboBoxes(self, row):
        """ Deletes all comboboxes of a row
        """
        tree = self.tree
        model = self.tree.model()

        for col in range(self.COL_FIRST_COMBO, self.maxCombos):
            logger.debug("Removing combobox at: ({}, {})".format(row, col))
            tree.setIndexWidget(model.index(row, col), None)

        self._comboBoxes = []


    @property
    def _hasFakeDimension(self):
        """ Returns True if the comboboxes have a fake dimension.

            We add a fake dimension (of length 1) only if there are fewer dimensions in the RTI
            than the inspector can show.
        """
        return self._rti.nDims < len(self._comboBoxes)


    # TODO: why again don't we create the comboxes when a new RTI is selected (just like the spins)?
    def _populateComboBoxes(self, row):
        """ Populates the combo boxes with values of the repo tree item
        """
        logger.debug("_populateComboBoxes")
        for comboBox in self._comboBoxes:
            comboBox.clear()

        if not self.rtiIsSliceable:
            # Add an empty item to the combo boxes so that resize to contents works.
            for comboBoxNr, comboBox in enumerate(self._comboBoxes):
                comboBox.addItem('', userData=None)
                comboBox.setEnabled(False)
            return

        nDims = self._rti.nDims
        nCombos = len(self._comboBoxes)

        for comboBoxNr, comboBox in enumerate(self._comboBoxes):

            # Only add a fake dimension of length 1 if there are fewer dimensions in the RTI than
            # the inspector can show.

            if self._hasFakeDimension:
                comboBox.addItem(FAKE_DIM_NAME, userData = FAKE_DIM_OFFSET + comboBoxNr)

            for dimNr in range(nDims):
                comboBox.addItem(self._rti.dimensionNames[dimNr], userData=dimNr)

            comboBox.setEnabled(True)
            #comboBox.adjustSize() # necessary?

        # Set combo-boxes current index
        if nDims >= nCombos:
            # If there are more dimensions in the RTI than the inspector can show, we use the
            # last couple of RTI dimensions. This because the NetCDF-CF conventions have the
            # preferred dimension order of T, Z, Y, X. On top of that, dimensions with length <= 1
            # are skipped if possible (see Github issue #9).

            curDim = nDims

            for comboBoxNr, comboBox in reversed(list(enumerate(self._comboBoxes))):
                while(True):
                    curDim -= 1
                    if curDim <= comboBoxNr or self.rti.arrayShape[curDim] > 1:
                        break

                curIdx = curDim
                assert 0 <= curIdx <= nDims, \
                    "curIdx should be <= {}, got {}".format(nDims, curIdx)

                comboBox.setCurrentIndex(curIdx)
        else:
            # If there are less dimensions in the RTI than the inspector can show, we fill
            # the combo-boxes starting at the leftmost and set the remaining comboboxes to the
            # fake dimension. This means that a table inspector will have one column and many
            # rows, which is the most convenient.

            for comboBoxNr, comboBox in enumerate(self._comboBoxes):
                curIdx = comboBoxNr + 1 if comboBoxNr < nDims else 0

                assert 0 <= curIdx <= nDims + 1, \
                    "curIdx should be <= {}, got {}".format(nDims + 1, curIdx)

                comboBox.setCurrentIndex(curIdx)



    # def getComboBoxDimensionName(self, comboBoxNr):
    #     """ Returns the currently selected dimension index of combo box nr comboBoxNr
    #     """
    #     self._comboBoxes[comboBoxNr].currentText()
    #
    #
    # def getComboBoxDimensionIndex(self, comboBoxNr):
    #     """ Returns the currently selected dimension index of combo box nr comboBoxNr
    #     """
    #     return self._comboBoxDimensionIndex(self._comboBoxes[comboBoxNr])


    def _comboBoxDimensionIndex(self, comboBox):
        """ Returns the dimension index (from the user data) of the current item of the combo box.
        """
        return comboBox.itemData(comboBox.currentIndex())


    def _dimensionSelectedInComboBox(self, dimNr):
        """ Returns True if the dimension is selected in one of the combo boxes.
        """
        for combobox in self._comboBoxes:
            if self._comboBoxDimensionIndex(combobox) == dimNr:
                return True
        return False


    def _createSpinBoxes(self, row):
        """ Creates a spinBox for each dimension that is not selected in a combo box.
        """
        assert len(self._spinBoxes) == 0, "Spinbox list not empty. Call _deleteSpinBoxes first"

        if not self.rtiIsSliceable:
            return

        logger.debug("_createSpinBoxes, array shape: {}".format(self._rti.arrayShape))

        self._setColumnCountForContents()

        tree = self.tree
        model = self.tree.model()
        col = self.COL_FIRST_COMBO + self.maxCombos

        for dimNr, dimSize in enumerate(self._rti.arrayShape):

            if self._dimensionSelectedInComboBox(dimNr):
                continue

            self._setHeaderLabel(col, '')

            spinBox = CollectorSpinBox()
            self._spinBoxes.append(spinBox)

            spinBox.setKeyboardTracking(False)
            spinBox.setCorrectionMode(QtWidgets.QAbstractSpinBox.CorrectToNearestValue)

            spinBox.setMinimum(0)
            spinBox.setMaximum(dimSize - 1)
            spinBox.setSingleStep(1)
            spinBox.setValue(dimSize // 2) # select the middle of the slice
            spinBox.setPrefix("{} = ".format(self._rti.dimensionNames[dimNr]))
            spinBox.setProperty("dim_nr", dimNr)
            #spinBox.adjustSize() # necessary?

            if not USE_SLIDER:
                spinBox.setSuffix("/{}".format(spinBox.maximum()))

            if dimSize == 1:
                spinBox.setEnabled(False)

            # This must be done after setValue to prevent emitting too many signals
            spinBox.valueChanged[int].connect(self._spinboxValueChanged)

            if USE_SLIDER:
                spinSlider = SpinSlider(spinBox, layoutContentsMargins = (0, 0, 5, 0))
                self._spinSliders.append(spinSlider)
                tree.setIndexWidget(model.index(row, col), spinSlider)
            else:
                tree.setIndexWidget(model.index(row, col), spinBox)

            col += 1

        # Resize the spinbox columns to their new contents
        self.tree.resizeColumnsFromContents(startCol=self.COL_FIRST_COMBO + self.maxCombos)


    def _deleteSpinBoxes(self, row):
        """ Removes all spinboxes
        """
        tree = self.tree
        model = self.tree.model()

        for col, spinBox in enumerate(self._spinBoxes, self.COL_FIRST_COMBO + self.maxCombos):
            spinBox.valueChanged[int].disconnect(self._spinboxValueChanged)
            tree.setIndexWidget(model.index(row, col), None)
        self._spinBoxes = []

        self._setColumnCountForContents()


    @QtSlot(int)
    def _comboBoxActivated(self, index, comboBox=None):
        """ Is called when a combo box value was changed by the user.

            Updates the spin boxes and sets other combo boxes having the same index to
            the fake dimension of length 1.
        """
        if comboBox is None:
            comboBox = self.sender()
        assert comboBox, "comboBox not defined and not the sender"

        blocked = self.blockChildrenSignals(True)

        curDimIdx = self._comboBoxDimensionIndex(comboBox)

        # if curDimIdx < FAKE_DIM_OFFSET:

        # Find any dimension that is not used
        unusedDimensions = list(range(self.rti.nDims + int(self._hasFakeDimension)))
        otherComboBoxes = [cb for cb in self._comboBoxes if cb is not comboBox]
        for otherComboBox in otherComboBoxes:
            if otherComboBox.currentIndex() in unusedDimensions:
                unusedDimensions.remove(otherComboBox.currentIndex())

        logger.info("Unused dimensions: {}".format(unusedDimensions))
        for otherComboBox in otherComboBoxes:
            if otherComboBox.currentIndex() == comboBox.currentIndex():
                otherComboBox.setCurrentIndex(unusedDimensions.pop())

        # Show only spin boxes that are not selected
        row = 0
        self._deleteSpinBoxes(row)
        self._createSpinBoxes(row)
        self._updateRtiInfo()

        self.blockChildrenSignals(blocked)

        logger.debug("{} sigContentsChanged signal (comboBox)"
                      .format("Blocked" if self.signalsBlocked() else "Emitting"))
        self.sigContentsChanged.emit(UpdateReason.COLLECTOR_COMBO_BOX)


    @QtSlot(int)
    def _spinboxValueChanged(self, index, spinBox=None):
        """ Is called when a spin box value was changed.
        """
        if spinBox is None:
            spinBox = self.sender()
        assert spinBox, "spinBox not defined and not the sender"

        self._updateRtiInfo()

        logger.debug("{} sigContentsChanged signal (spinBox)"
                      .format("Blocked" if self.signalsBlocked() else "Emitting"))
        self.sigContentsChanged.emit(UpdateReason.COLLECTOR_SPIN_BOX)


    def getSlicedArray(self, copy=True):
        """ Slice the rti using a tuple of slices made from the values of the combo and spin boxes.

            :param copy: If True (the default), a copy is made so that inspectors cannot
                accidentally modify the underlying of the RTIs. You can set copy=False as a
                potential optimization, but only if you are absolutely sure that you don't modify
                the the slicedArray in your inspector! Note that this function calls transpose,
                which can still make a copy of the array for certain permutations.

            :return: ArrayWithMask array with the same number of dimension as the number of
                comboboxes (this can be zero!).

                Returns None if no slice can be made (i.e. the RTI is not sliceable).
            :rtype ArrayWithMask:
        """
        #logger.debug("getSlicedArray() called")

        if not self._rti:
            self.sigShowMessage.emit("No item selected.")
            return None

        if not self.rtiIsSliceable:
            # This is very common so we don't show a message so the user isn't flooded.
            # Also this can have many different causes (compound data, lists, etc_ so it's
            # difficult to show a good, descriptive message.
            return None

        if np.prod(self._rti.arrayShape) == 0:
            self.sigShowMessage.emit("Selected item has zero array elements.")
            return None

        # The dimensions that are selected in the combo boxes will be set to slice(None),
        # the values from the spin boxes will be set as a single integer value
        nDims = self.rti.nDims
        sliceList = [slice(None)] * nDims

        for spinBox in self._spinBoxes:
            dimNr = spinBox.property("dim_nr")
            sliceList[dimNr] = spinBox.value()

        # Make the array slicer. It needs to be a tuple, a list of only integers will be
        # interpreted as an index. With a tuple, array[(exp1, exp2, ..., expN)] is equivalent to
        # array[exp1, exp2, ..., expN].
        # See: http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html
        slicedArray = self.rti[tuple(sliceList)]

        # Make a copy to prevent inspectors from modifying the underlying array.
        if copy:
            if versionStrToTuple(np.__version__) >= (1,19,0):
                slicedArray = np.copy(slicedArray, subok=True)  # Fixes issue #8
            else:
                slicedArray = ma.copy(slicedArray)

        # If there are no comboboxes the sliceList will contain no Slices objects, only ints. Then
        # the resulting slicedArray will be a usually a scalar (only structured fields may yield an
        # array). We convert this scalar to a zero-dimensional Numpy array so that inspectors
        # always get an array (having the same number of dimensions as the dimensionality of the
        # inspector, i.e. the number of comboboxes).
        # Also scalar RTIs, which have nDim == 0, can return a scalar which must be converted.
        # TODO: perhaps always convert to array.
        if self.maxCombos == 0 or self.rti.nDims == 0:
            slicedArray = ma.MaskedArray(slicedArray)

        # Post-condition type check
        check_is_an_array(slicedArray, np.ndarray)

        # Enforce the return type to be a masked array.
        if not isinstance(slicedArray, ma.MaskedArray):
            slicedArray = ma.MaskedArray(slicedArray)

        # Add fake dimensions of length 1 so that result.ndim will equal the number of combo boxes
        # TODO: Perhaps get rid of this because it fails with masked arrays with fill values.
        # The less we do here, the less chance an error occurs. See development/todo.txt
        # 2022-02-21: this seems no longer a problem (numpy 1.22.2)
        for dimNr in range(slicedArray.ndim, self.maxCombos):
            #logger.debug("Adding fake dimension: {}".format(dimNr))
            slicedArray = ma.expand_dims(slicedArray, dimNr)

        # Post-condition dimension check
        assert slicedArray.ndim == self.maxCombos, \
            "Bug: getSlicedArray should return a {:d}D array, got: {}D" \
            .format(self.maxCombos, slicedArray.ndim)

        # Convert to ArrayWithMask class for working around issues with the numpy maskedarray
        awm = ArrayWithMask.createFromMaskedArray(slicedArray)
        del slicedArray

        # Shuffle the dimensions to be in the order as specified by the combo boxes
        comboDims = [self._comboBoxDimensionIndex(cb) for cb in self._comboBoxes]
        permutations = np.argsort(comboDims)
        logger.debug("slicedArray.shape: {}".format(awm.data.shape))
        logger.debug("Transposing dimensions: {}".format(permutations))
        awm = awm.transpose(permutations)

        awm.checkIsConsistent()

        return awm


    def getSlicesString(self):
        """ Returns a string representation of the slices that are used to get the sliced array.
            For example returns '[:, 5]' if the combo box selects dimension 0 and the spin box 5.
        """
        if not self.rtiIsSliceable:
            return ''

        # The dimensions that are selected in the combo boxes will be set to slice(None),
        # the values from the spin boxes will be set as a single integer value
        nDims = self.rti.nDims
        sliceList = [':'] * nDims

        for spinBox in self._spinBoxes:
            dimNr = spinBox.property("dim_nr")
            sliceList[dimNr] = str(spinBox.value())

        # No need to shuffle combobox dimensions like in getSlicedArray; all combobox dimensions
        # yield a colon.

        return "[" + ", ".join(sliceList) + "]"


    # def independentDimensionNames(self):
    #     """ Returns list of the names of the independent dimensions, which have been selected in
    #         the combo boxes.
    #     """
    #     return [combobox.currentText() for combobox in self._comboBoxes]
    #
    #
    # def independentDimensionUnits(self):
    #     """ Returns list of the units of the independent dimensions, which have been selected in
    #         the combo boxes. At the moment return '' for each independent dimension.
    #     """
    #     return ['' for _combobox in self._comboBoxes]
    #
    #
    # def dependentDimensionName(self):
    #     """ Returns the name of the dependent dimension.
    #         A good default is the name of the RTI.
    #     """
    #     return self.rti.nodeName if self.rti else ''


    @property
    def rtiInfo(self):
        """ Returns a dictionary with information on the selected RTI (repo tree item).
            This can be used in string formatting of config options. For instance: the plot title
            can be specified as: '{path} {slices}', which will be expanded with the actual nodePath
            and slices-string of the RTI.

            If no RTI is selected the applicable values will be empty strings.

            The dictionary has the following contents:
                slices : a string representation of the selected slice indices.
                name        : nodeName of the RTI
                path        : nodePath of the RTI
                unit        : unit of the RTI in parentheses
                raw-unit    : unit of the RTI without parentheses (empty string if no unit given)
                n-dim       : dimension selected in the combobox for axis n. The axis name will be
                              in lower case (so e.g. x-dim, y-dim, etc)
        """
        return self._rtiInfo


    def _updateRtiInfo(self):
        """ Updates the _rtiInfo property when a new RTI is set or spin/comboboxes change value .
        """
        logger.debug("Updating self._rtiInfo")
        # Info about the dependent dimension
        rti = self.rti
        if rti is None:
            info = {'slices': '',
                    'name': '',
                    'path': '',
                    'file-name': '',
                    'dir-name': '',
                    'base-name': '',
                    'unit': '',
                    'raw-unit': ''}
        else:
            dirName, baseName = os.path.split(rti.fileName)
            info = {'slices': self.getSlicesString(),
                    'name': rti.nodeName,
                    'path': rti.nodePath,
                    'file-name': rti.fileName,
                    'dir-name': dirName,
                    'base-name': baseName,
                    'unit': '({})'.format(rti.unit) if rti.unit else '',
                    'raw-unit': rti.unit}

        # Add the info of the independent dimensions (appended with the axis name of that dim).
        for axisName, comboBox in zip(self._axisNames, self._comboBoxes):
            dimName = comboBox.currentText()
            key = '{}-dim'.format(axisName.lower())
            info[key] = dimName

        self._rtiInfo = info




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

# To make a release follow these steps:
#   python setup.py sdist
#   twine upload dist/argos-0.2.0rc1.tar.gz
# or better
#   rm -rf build dist
#   python setup.py bdist_wheel
#   twine check dist/*
#   twine upload dist/argos-x.y.z-py3-none-any.whl

# If you get invalid command 'bdist_wheel', you must install the 'wheel' package first.

# See also https://packaging.python.org/en/latest/distributing.html
# TODO: still can't make a wheel even following the instructions in the link below.
# http://stackoverflow.com/questions/26664102/why-can-i-not-create-a-wheel-in-pyt

import os
import sys

def err(*args, **kwargs):
    sys.stderr.write(*args, **kwargs)
    sys.stderr.write('\n')


try:
    from setuptools import setup, find_packages
except ImportError:
    err("Argos requires setuptools for intallation. (https://pythonhosted.org/an_example_pypi_project/setuptools.html)")
    err("You can download and install it simply with: python argos/external/ez_setup.py")
    sys.exit(1)


from argos import info

if sys.version_info < (3,7):
    err("Argos requires Python 3.7")
    sys.exit(1)


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')


# Don't require PyQt for two reasons. First users may use PySide2 (although at the moment PySide is
# not yet working). Second, users may use Anaconda to install PyQt. Anaconda uses a different
# package name (pyqt) than pip (PyQt5) and the tools can't detect correctly if PyQt has been
# installed. This leads to trouble. See:
#   https://www.anaconda.com/using-pip-in-a-conda-environment/
#   https://github.com/ContinuumIO/anaconda-issues/issues/1554
#
# In Debian pip will ignored installed system packages (e.g. use --ignore-installed).
# To override this use: export PIP_IGNORE_INSTALLED=0
# See https://github.com/pypa/pip/issues/4222

install_requires = [
    #"PyQt5 >= 5.6.0", # Don't require PyQt. See comment above
    "cmlib >= 1.1.2",  # Needed, even if no plugins are installed.
    "numpy >= 1.11",
    # Argos will technically work without pyqtgraph and h5py, but with very limited functionality.
    "pgcolorbar >= 1.1.1",
    "pyqtgraph >= 0.11",
    # "h5py >= 2.6"
]

long_description = readme + '\n\n' + history
print(long_description)

setup(
    name = info.REPO_NAME,
    version = info.VERSION,
    description = info.SHORT_DESCRIPTION,
    long_description = readme + '\n\n' + history,
    long_description_content_type = 'text/x-rst',
    author = info.AUTHOR,
    author_email = info.EMAIL,
    license = "GPLv3",
    url=info.PROJECT_URL,
    packages = find_packages(),
    package_data = {'': ['HISTORY.rst'],
                    info.PACKAGE_NAME: ['img/argos.css', 'img/snipicons/*', 'utils/default_logging.json']},
    entry_points={'gui_scripts': ['argosw = argos.main:main'],
                  'console_scripts': ['argos = argos.main:main',
                                      'argos_make_wrappers = argos.argos_make_wrappers:main']},
    install_requires = install_requires,
    zip_safe = False,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities',
    ],
    keywords = 'NetCDF HDF5 plotting graphs',
    #test_suite='tests',
    #tests_require=test_requirements
)



# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" The table model classes that form the base of the registries
"""

import logging

from argos.utils.cls import type_name, check_class

from argos.qt import QtCore, QtSignal, Qt


logger = logging.getLogger(__name__)


class BaseItem(object):
    """ An object that is stored in the BaseItemStore.

        It always must have a name, which is used as the identifier
    """
    FIELDS = []  # The fields that this item contains.
    LABELS = []  # Human readable label for each field. Should be overridden in descendants
    STRETCH = [] # True -> the corresponding table column can stretch. False -> resize to contents.

    # For debugging...
    # FIELDS = ['debugCount', 'name', 'path']
    # LABELS = FIELDS # For debugging.
    # STRETCH = [True, False, True]

    COL_DECORATION = None   # Column number that contains the decoration. None for no icons

    _sequenceCounter = 0  # For debugging

    def __init__(self, **kwargs):
        """ Constructor
        """
        self._data = {}

        # Set fields from keyword args
        for key, value in kwargs.items():
            if key not in self.FIELDS:
                raise ValueError("Key '{}' not in field names: {}".format(key, self.FIELDS))
            self._data[key] = value

        # Set remaining fields to empty string
        for key in self.FIELDS:
            if key not in self._data:
                if key == 'debugCount':
                    self._data['debugCount'] = BaseItem._sequenceCounter
                    BaseItem._sequenceCounter += 1
                else:
                    self._data[key] = ''

        assert len(self.FIELDS) > 0
        if len(self.LABELS) != len(self.FIELDS):
            raise AssertionError("Number of labels ({}) is not equal to number of fields ({})"
                                 .format(len(self.LABELS), len(self.FIELDS)))

        if len(self.STRETCH) != len(self.FIELDS):
            raise AssertionError("Number of stretches ({}) is not equal to number of labels ({})"
                                 .format(len(self.STRETCH), len(self.FIELDS)))



    def __repr__(self):
        return "<{}: {}>".format(type_name(self), self._data)


    @property
    def data(self):
        """ The data dictionary
        """
        return self._data


    @property
    def decoration(self):
        """ A optional icon that is displayed in the COL_DECORATION column.

            The base implementation returns None (no icon). Descendants can override this
        """
        return None


    #####
    # The follow functions load or save their state to JSON config files.
    #####

    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        cfg = {}
        for field in self.FIELDS:
            cfg[field] = str(self._data[field])

        return cfg


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        self._fields = {}
        for field in self.FIELDS:
            if field in cfg:
                self._data[field] = cfg[field]
            else:
                logger.warning("Field '{}' not in config: {}".format(field, cfg))



class BaseItemStore(object):
    """ Class that stores a collection of BaseItems or descendants. Base class for the registries.

        In principle this class could be merged with the BaseItemModel but I chose to separate them
        because most of the time the data is used read-only, from the store. Only when the user
        edits the registry from the GUI is a table model needed. This way registries don't descent
        from QAbstractTableModel and inherit a huge number of methods.

        The BaseItemStore can only store items of one type (ITEM_CLASS). Descendants will
        store their own type. For instance the InspectorRegistry will store InspectorstoreItem
        items. This makes serialization easier.
    """

    # All items in this store will be of this class. BaseModel descendant will typically override
    # the ITEM_CLASS value
    ITEM_CLASS = BaseItem

    def __init__(self):
        """ Constructor
        """
        self._items = []


    def __str__(self):
        return "Item store"

    @property
    def fieldNames(self):
        """ Name of the fields. So think twice before changing them."""
        return self.ITEM_CLASS.FIELDS


    @property
    def fieldLabels(self):
        """ Short, human readable, label fields for use in GUI. """
        return self.ITEM_CLASS.LABELS


    @property
    def canStretchPerColumn(self):
        """ List of booleans, for each column indicating if it can strech.
            True -> the corresponding table column can stretch. False -> resize to contents.
        """
        return self.ITEM_CLASS.STRETCH


    @property
    def items(self):
        """ The registered class items.
        """
        return self._items


    def clear(self):
        """ Empties the registry
        """
        self._items = []


    #####
    # The follow functions load or save their state to JSON config files.
    #####

    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        return [item.marshall() for item in self.items]


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        self.clear()
        if not cfg:
            logger.info("Empty config, using registry defaults for: {}".format(self))
            for storeItem in self.getDefaultItems():
                self._items.append(storeItem)
        else:
            for dct in cfg:
                logger.debug("Creating {} from: {}".format(self.ITEM_CLASS, dct))
                storeItem = self.ITEM_CLASS()
                storeItem.unmarshall(dct)
                self._items.append(storeItem)


    def getDefaultItems(self):
        """ Returns a list with the default items.
            This is used initialize the application plugins when there are no saved settings,
            for instance the first time the application is started.
            The base implementation returns an empty list but other registries should override it.
        """
        return []



# Rhe main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201


class BaseTableModel(QtCore.QAbstractTableModel):

    sigItemChanged = QtSignal(BaseItem) # An item has changed.

    def __init__(self, store, parent=None):
        """ Constructor.

            :param store: Underlying data store, must descent from BaseItemStore
            :param parent: Parent widget
        """
        super(BaseTableModel, self).__init__(parent)
        check_class(store, BaseItemStore)
        self._store = store
        self._fieldNames = self._store.fieldNames
        self._fieldLabels = self.store.fieldLabels

    @property
    def store(self):
        """ The underlying BaseItemStore

            Note that if you modify this directly (i.e. not via the setData method), the model is
            not aware of the modifications. Therefore the user is responsible for nofifying the
            this table model after modifications.
        """
        return self._store


    def rowCount(self, parent=None):
        """ Returns the number of items in the registry."""
        return len(self._store.items)


    def columnCount(self, parent=None):
        """ Returns the number of columns of the registry."""
        return len(self._fieldNames)


    def itemFromIndex(self, index, altItem=None):
        """ Gets the item given the model index
            Returns altItem (default: None) if the index is not alid
        """
        if not index.isValid():
            return altItem
        else:
            return self._store.items[index.row()]


    def indexFromItem(self, storeItem, col=0):
        """ Gets the index (with column=0) for the row that contains the storeItem
            If col is negative, it is counted from the end
        """
        if col < 0:
            col = len(self._fieldNames) - col
        try:
            row = self._store.items.index(storeItem)
        except ValueError:
            return QtCore.QModelIndex()
        else:
            return self.index(row, col)


    def flags(self, index):
        """ Returns the item flags for the given index.
        """
        return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable


    def headerData(self, section, orientation, role):
        """ Returns the header for a section (row or column depending on orientation).
            Reimplemented from QAbstractTableModel to make the headers start at 0.
        """
        if role == QtCore.Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self._fieldLabels[section]
            else:
                return str(section)
        else:
            return None


    def data(self, index, role=Qt.DisplayRole):
        """ Returns the data stored under the given role for the item referred to by the index.
        """
        if not index.isValid():
            return None

        if role not in (Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole, Qt.DecorationRole):
            return None

        row = index.row()
        col = index.column()
        item = self._store.items[row]
        attrName = self._fieldNames[col]

        if role in (Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole):
            return str(item.data[attrName])

        elif role == Qt.DecorationRole:
            if col == item.COL_DECORATION:
                return item.decoration

        # elif role == Qt.ForegroundRole:
        #     return self.regularBrush
        else:
            raise ValueError("Invalid role: {}".format(role))


    def setData(self, index, value, role=Qt.EditRole):
        """ Sets the role data for the item at index to value.
        """
        if not index.isValid():
            return False

        if role != Qt.EditRole:
            return False

        row = index.row()
        col = index.column()
        storeItem = self._store.items[row]
        fieldName = self._fieldNames[col]
        storeItem.data[fieldName] = value

        self.emitDataChanged(storeItem)

        return True


    def emitDataChanged(self, storeItem):
        """ Emits the dataChanged and sigItemChanged signals for the storeItem
        """
        leftIndex = self.indexFromItem(storeItem, col=0)
        rightIndex = self.indexFromItem(storeItem, col=-1)

        logger.debug("Data changed: {} ... {}".format(self.data(leftIndex), self.data(rightIndex)))
        self.dataChanged.emit(leftIndex, rightIndex)
        self.sigItemChanged.emit(storeItem)


    def createItem(self):
        """ Creates an emtpy item of type ITEM_CLASS
        """
        return self.store.ITEM_CLASS()


    def insertItem(self, item, row):
        """ Insert an item in the store at a certain row.
        """
        check_class(item, self.store.ITEM_CLASS)
        logger.info("Inserting {!r} at row {}".format(item, row, self))
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        try:
            self.store.items.insert(row, item)
        finally:
            self.endInsertRows()


    def popItemAtRow(self, row):
        """ Removes a store item from the store.
            Returns the item
        """
        self.beginRemoveRows(QtCore.QModelIndex(), row, row)
        try:
            item = self.store.items[row]
            del self.store.items[row]
            return item
        finally:
            self.endRemoveRows()


    def moveItem(self, fromRow, toRow):
        """ Moves the item for position
        """
        item = self.popItemAtRow(fromRow)

        # This always works, regardless if fromPos is before or after toPos
        self.insertItem(item, toRow)




def main():
    """ Test classes
    """
    from pprint import pprint

    store2 = BaseItemStore()
    store2.unmarshall("[{'name': '0'}, {'name': 'A'},  {'name': 'b'},  {'name': 'b'},  "
                      "{'name': 'B'},  {'name': 'C'}]")

    print("Store2")
    pprint(store2)




if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the TestWalkDialog widget.
"""
import base64
import copy
import logging
import time
from typing import Optional, Tuple, List

from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSlot
from argos.qt.misc import getWidgetGeom, getWidgetState
from argos.widgets.constants import MONO_FONT, FONT_SIZE, COLOR_ERROR
from argos.widgets.misc import processEvents
from argos.utils.cls import check_is_a_sequence
from argos.utils.config import ConfigDict
from argos.utils.misc import wrapHtmlColor

logger = logging.getLogger(__name__)


# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201


class TestWalkDialog(QtWidgets.QDialog):
    """ Dialog that shows progress of test walk through the data and gives a summary when done.
    """
    def __init__(self, mainWindow, parent=None):
        """ Constructor

            Args:
                mainWindow: the Argos main window.
        """
        super(TestWalkDialog, self).__init__(parent=parent)
        self.setModal(False)

        self._isOngoing = False
        self._showAllResults = False

        self._mainWindow = mainWindow

        self._currentTestName: Optional[str] = None
        self._results: List[Tuple[bool, str]] = []

        self.walkCurrentAction = QtWidgets.QAction("Walk Current Item", self)
        self.walkCurrentAction.setToolTip("Does a test walk on the currently selected tree item.")
        self.walkCurrentAction.triggered.connect(self.walkCurrentRepoNode)
        self.addAction(self.walkCurrentAction)

        self.walkAllAction = QtWidgets.QAction("Walk All Items", self)
        self.walkAllAction.setToolTip("Does a test walk on all tree nodes.")
        self.walkAllAction.triggered.connect(self.walkAllRepoNodes)
        self.addAction(self.walkAllAction)

        self.abortWalkAction = QtWidgets.QAction("Abort Walk", self)
        self.abortWalkAction.setToolTip("Aborts the current test walk.")
        self.abortWalkAction.triggered.connect(self.abortTestWalk)
        self.addAction(self.abortWalkAction)

        #################
        # Setup widgets #
        #################

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.mainLayout)

        self.controlLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.addLayout(self.controlLayout)

        self.walkCurrentButton = QtWidgets.QToolButton()
        self.walkCurrentButton.setDefaultAction(self.walkCurrentAction)
        self.controlLayout.addWidget(self.walkCurrentButton)

        self.walkAllButton = QtWidgets.QToolButton()
        self.walkAllButton.setDefaultAction(self.walkAllAction)
        self.controlLayout.addWidget(self.walkAllButton)

        self.abortWalkButton = QtWidgets.QToolButton()
        self.abortWalkButton.setDefaultAction(self.abortWalkAction)
        self.controlLayout.addWidget(self.abortWalkButton)

        self.allInspectorsCheckBox = QtWidgets.QCheckBox("Test all Inspectors")
        self.controlLayout.addWidget(self.allInspectorsCheckBox)
        self.allDetailTabsCheckBox = QtWidgets.QCheckBox("Test all Detail-tabs")
        self.controlLayout.addWidget(self.allDetailTabsCheckBox)
        self.controlLayout.addStretch()

        self.progressBar = QtWidgets.QProgressBar()
        self.mainLayout.addWidget(self.progressBar)

        self.curPathLabel = QtWidgets.QLabel()
        self.curPathLabel.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.mainLayout.addWidget(self.curPathLabel)

        font = QtGui.QFont()
        font.setFamily(MONO_FONT)
        font.setFixedPitch(True)
        font.setPointSize(FONT_SIZE)

        self.editor = QtWidgets.QPlainTextEdit()
        self.editor.setReadOnly(True)
        self.editor.setFont(font)
        self.editor.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.editor.clear()
        self.mainLayout.addWidget(self.editor)

        self.bottomLayout = QtWidgets.QHBoxLayout()
        self.bottomLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.addLayout(self.bottomLayout)

        self.showAllResCheckBox = QtWidgets.QCheckBox("Show all results")
        self.showAllResCheckBox.stateChanged.connect(self.onShowAllResultsChanged)
        self.bottomLayout.addWidget(self.showAllResCheckBox)

        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Close)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        self.bottomLayout.addWidget(buttonBox)

        self.resize(QtCore.QSize(800, 400))
        self._updateButtons()


    def finalize(self):
        """ Is called before destruction (when closing).
            Can be used to clean-up resources.
        """
        self.showAllResCheckBox.stateChanged.disconnect(self.onShowAllResultsChanged)


    @property
    def results(self):
        """ Get a copy of the test results
        """
        return copy.deepcopy(self._results)


    @property
    def repoWidget(self):
        """ The repoWidget of the main window
        """
        return self._mainWindow.repoWidget


    @property
    def repoTreeView(self):
        """ The repoTreeView of repo widget of the main window
        """
        return self._mainWindow.repoWidget.repoTreeView


    def marshall(self) -> Tuple[ConfigDict, ConfigDict]:
        """ Returns a layout and state config dictionaries
        """
        layoutCfg = dict(
            winGeom = base64.b64encode(getWidgetGeom(self)).decode('ascii'),
        )

        cfg = dict(
            testAllInspectors = self.allInspectorsCheckBox.isChecked(),
            testAllDetailTabs = self.allDetailTabsCheckBox.isChecked(),
            showAllResults = self.showAllResCheckBox.isChecked(),
        )
        return layoutCfg, cfg


    def unmarshall(self, layoutCfg, cfg):
        """ Initializes itself from a layout and config dict form the persistent settings.
        """
        if 'testAllInspectors' in cfg:
            self.allInspectorsCheckBox.setChecked(cfg['testAllInspectors'])

        if 'testAllDetailTabs' in cfg:
            self.allDetailTabsCheckBox.setChecked(cfg['testAllDetailTabs'])

        if 'showAllResults' in cfg:
            self.showAllResCheckBox.setChecked(cfg['showAllResults'])

        if 'winGeom' in layoutCfg:
            self.restoreGeometry(base64.b64decode(layoutCfg['winGeom']))


    def reject(self):
        """ Called when the user closes the dialog. Aborts any running test walk.
        """
        logger.debug("Closing TestWalkDialog")
        self.abortTestWalk()
        super().reject()


    def _updateButtons(self):
        """ Enables/disables buttons depending on if the test is ongaing
        """
        self.walkAllButton.setEnabled(not self._isOngoing)
        self.walkCurrentButton.setEnabled(not self._isOngoing)
        self.abortWalkButton.setEnabled(self._isOngoing)


    def _setProgressFraction(self, fraction: float):
        """ Sets the fraction (percentage / 100)
        """
        self.progressBar.setValue(int(round(fraction * 100)))
        logger.debug("Progress fraction {:8.3f}".format(fraction))
        processEvents()


    @QtSlot(int)
    def onShowAllResultsChanged(self, state: int):
        """ Executed when the 'show all results' check box changes
        """
        logger.info("Setting onShowAllResultsChanged: {}".format(state))
        if state == Qt.Checked:
            self._showAllResults = True
        elif state == Qt.Unchecked:
            self._showAllResults = False
        else:
            raise ValueError("Unexpected checkbox state: {}".format(state))


    def appendText(self, text: str, isError: bool = False):
        """ Appends a text message to the editor.
        """
        if not isError:
            self.editor.appendPlainText(text)
        else:
            self.editor.appendHtml(wrapHtmlColor(text, COLOR_ERROR))


    @QtSlot(bool)
    def setTestResult(self, success: bool):
        """ Appends the currently selected path node to the list of tests.

            This slot will be called whenever the inspector or detail tab (properties, attributes,
            quicklook) updates itself.

            Setting the name and result are separate methods because the inspector and detail tab
            don't know which node is currently selected in the repo tree.
        """
        if not self._isOngoing:
            logger.debug("No test ongoing. Test result discarded.")
            return

        if self._currentTestName is None:
            # When we test also for different inspector or detail panels, the test with the
            # current inspector & panels is ignored because the same test will be repeated later
            # This prevents showing the same test result again but with a different name.
            logger.debug("Ignoring test with current inspector and detail panel")
            return

        # An inspector may be updated twice during a single test. For example, if a node was not
        # yet expanded, expanding will redraw the inspector for that node. In that case this
        # function will be called twice for the current test. We therefore merge these results.
        if self._results:
            prevSuccess, prevName = self._results[-1]
            if self._currentTestName == prevName:
                logger.debug("Ignoring duplicate test result: {}".format(self._currentTestName))
                # It can happen that test results differ. For instance when the inspector fails
                # but the detail panel succeeds.
                self._results[-1] = (success and prevSuccess, self._currentTestName)
                return

        self._results.append((success, self._currentTestName))

        if self._showAllResults or not success:
            line = "{:8s}: {}".format("success" if success else "FAILED", self._currentTestName)
            logger.info("setTestResult: {}".format(line))
            self.appendText(line, isError=not success)


    @QtSlot()
    def abortTestWalk(self):
        """ Sets the flag to abort the test walk
        """
        self._isOngoing = False
        self._updateButtons()
        self.curPathLabel.setText("Test walk aborted!")


    @QtSlot()
    def walkCurrentRepoNode(self, allInspectors: bool=None, allDetailTabs: bool=None):
        """ Will visit all nodes below the currently selected node
        """
        logger.debug(f"walkCurrentRepoNode: {allInspectors} {allDetailTabs}")
        if allInspectors is not None:
            self.allInspectorsCheckBox.setChecked(allInspectors)

        if allDetailTabs is not None:
            self.allDetailTabsCheckBox.setChecked(allDetailTabs)

        curItem, _curIdx = self.repoTreeView.getCurrentItem()
        if curItem is None:
            logger.warning("No node selected for test walk")
        else:
            logger.info("Test walk current item: {}".format(curItem.nodePath))
            self.walkRepoNodes([curItem.nodePath])


    @QtSlot()
    def walkAllRepoNodes(self, allInspectors: bool=None, allDetailTabs: bool=None):
        """ Will visit all nodes in the repo tree.

            See walkRepoNodes docstring for more info
        """
        logger.info("testWalkAllNodes")
        if allInspectors is not None:
            self.allInspectorsCheckBox.setChecked(allInspectors)

        if allDetailTabs is not None:
            self.allDetailTabsCheckBox.setChecked(allDetailTabs)

        repo = self.repoTreeView.model()
        nodePaths = [rti.nodePath for rti in repo.rootItems()]
        logger.debug("All root items: {}".format(nodePaths))
        self.walkRepoNodes(nodePaths)


    def walkRepoNodes(self, nodePaths):
        """ Will recursively walk through a list of repo tree nodes and all their descendants

            Is useful for testing.
        """
        logger.info("-------------- Running Tests ----------------")
        logger.debug("Visiting all nodes below: {}".format(nodePaths))

        self.show()
        self._currentTestName = None
        self._results = []
        self.editor.clear()

        nodesVisited = 0
        logger.debug("Starting test walk")
        self._isOngoing = True
        self._updateButtons()

        # Unselect the current item to force the first node to trigger a new inspector
        originalIndex = self.repoTreeView.currentIndex()
        invalidIndex = self.repoTreeView.model().index(-1, -1)
        assert not invalidIndex.isValid(), "sanity check"
        self.repoTreeView.setCurrentIndex(invalidIndex)

        # Similarly, select the last detail panel to force the first to trigger a new result
        self.repoWidget.tabWidget.setCurrentIndex(self.repoWidget.tabWidget.count()-1)

        try:
            timeAtStart = time.perf_counter()
            check_is_a_sequence(nodePaths) # prevent accidental iteration over strings.

            for numVisited, nodePath in enumerate(nodePaths):
                nodeItem, nodeIndex = self.repoTreeView.model().findItemAndIndex(nodePath)

                assert nodeItem is not None, "Test data not found, rootNode: {}".format(nodePath)
                assert nodeIndex

                progressRange = (numVisited / len(nodePaths), (numVisited+1) / len(nodePaths))
                nodesVisited += self._visitNodes(nodeIndex, progressRange)

            if self._isOngoing:
                self._setProgressFraction(1.0)

            duration = time.perf_counter() - timeAtStart
            self._logTestSummary(duration, nodesVisited)
            self._displayTestSummary(duration, nodesVisited)
        finally:
            logger.debug("Stopping test walk")
            self._isOngoing = False
            self._currentTestName = None
            self._updateButtons()
            if self._isOngoing:
                self.repoTreeView.setCurrentIndex(originalIndex)
            logger.info("-------------- Test walk done ----------------")


    def _visitNodes(self, index: QtCore.QModelIndex, progressRange: Tuple[float, float]) -> int:
        """ Helper function that visits all the nodes recursively.

            Args:
                index: current node index
                progressRange: range that this call will cover in the progress bar.

            Returns:
                The number of nodes that where visited.
        """
        assert index.isValid(), "sanity check"

        if not self._isOngoing:
            return 0  # Test walk aborted.

        repoModel = self.repoTreeView.model()
        item = repoModel.getItem(index)
        logger.info("Visiting: {!r} ({} children)".
                    format(item.nodePath, repoModel.rowCount(index)))

        self.curPathLabel.setText(item.nodePath)

        # Select index
        if item.nodeName.startswith('_'):
            logger.warning("Skipping node during testing: {}".format(item.nodePath))
            return 0
        else:
            logger.debug("Not skipping node during testing: {}".format(item.nodePath))

        if self.allDetailTabsCheckBox.isChecked() or self.allInspectorsCheckBox.isChecked():
            self._currentTestName = None
        else:
            self._currentTestName = "{}".format(item.nodePath)
            processEvents()

        wasOpen = self.repoTreeView.isExpanded(index)
        self.repoTreeView.setCurrentIndex(index)
        self.repoTreeView.expand(index)

        if self.allDetailTabsCheckBox.isChecked():
            # Try properties, attributes and quicklook tabs
            for idx in range(self.repoWidget.tabWidget.count()):
                tabName = self.repoWidget.tabWidget.tabText(idx)
                self._currentTestName = "{:11}: {}".format(tabName, item.nodePath)
                logger.debug("Setting repo detail tab : {}".format(tabName))
                self.repoWidget.tabWidget.setCurrentIndex(idx)
                processEvents()

        if self.allInspectorsCheckBox.isChecked():
            for action in self._mainWindow.inspectorActionGroup.actions():
                assert action.text(), "Action text undefined: {!r}".format(action.text())
                self._currentTestName = "{:11}: {}".format(action.text(), item.nodePath)
                action.trigger()
                processEvents()

        nodesVisited = 1

        prMin, prMax = progressRange
        prLength = prMax - prMin
        self._setProgressFraction(prMin)
        logger.debug("Range {:8.3f}, {:8.3f}: {}".format(prMin, prMax, item.nodePath))

        toVisit = repoModel.rowCount(index)
        for numVisited, rowNr in enumerate(range(toVisit)):
            childIndex = repoModel.index(rowNr, 0, parentIndex=index)
            subRange = (prMin + numVisited / toVisit * prLength,
                        prMin + (numVisited+1) / toVisit * prLength)
            nodesVisited += self._visitNodes(childIndex, subRange)

        if self._isOngoing and not wasOpen:
            self.repoTreeView.closeItem(index)

        return nodesVisited


    def _displayTestSummary(self, duration: float, nodesVisited: int):
        """ Displays a test summary in the dialog
        """
        self.appendText('-' * 80)
        if self._isOngoing:
            self.appendText("Test finished.")
        else:
            self.appendText("Test ABORTED", isError=True)


    def _logTestSummary(self, duration: float, nodesVisited: int):
        """ Logs a summary of the test results.
        """
        logger.info("Visited {} nodes in {:.1f} seconds ({:.1f} nodes/second)."
                    .format(nodesVisited, duration, nodesVisited/duration))

        failedTests = [(success, name) for success, name in self._results if not success]
        logger.info("Number of failed tests during test walk: {}".format(len(failedTests)))
        for testName in failedTests:
            logger.info("    {}".format(testName))

        if not self._isOngoing:
            logger.info("")
            logger.info("NOTE: the test walk was aborted!")
        else:
            # Check that the number of test results is as expected
            allInspect = self.allInspectorsCheckBox.isChecked()
            allDetail = self.allDetailTabsCheckBox.isChecked()
            if allInspect or allDetail:
                numInspectors = len(self._mainWindow.inspectorActionGroup.actions())
                numDetailPanels = self.repoWidget.tabWidget.count()
                factor = numInspectors * int(allInspect) + numDetailPanels * int(allDetail)
            else:
                factor = 1

            expectedNumResuls = nodesVisited * factor
            if len(self._results) != expectedNumResuls:
                msg = "Actual nr of results ({}) != expected nr of results: {}"\
                    .format(len(self._results), expectedNumResuls)
                logger.warning(msg)

                # No assert, as this can happen when there are two nodes with the same name, which
                # is common in NetCDF files that follow the CF conventions.
                #assert False, msg



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Data repository functionality
"""
import logging
from argos.qt import Qt, QtCore
from argos.qt.treemodels import BaseTreeModel
#from argos.info import DEBUGGING
from argos.repo.filesytemrtis import createRtiFromFileName
from argos.repo.baserti import BaseRti
from argos.repo.registry import ICON_COLOR_UNDEF, RtiRegItem
from argos.utils.cls import to_string, type_name, check_class
from argos.utils.dirs import normRealPath

# The cast to int is necessary to avoid a bug in PySide, See:
# https://bugreports.qt-project.org/browse/PYSIDE-20
ALIGN_LEFT  = int(Qt.AlignVCenter | Qt.AlignLeft)
ALIGN_RIGHT = int(Qt.AlignVCenter | Qt.AlignRight)


logger = logging.getLogger(__name__)

class RepoTreeModel(BaseTreeModel):
    """ An implementation QAbstractItemModel that offers read-only access of the application data
        for QTreeViews. The underlying data is stored as repository tree items (BaseRti
        descendants).
    """
    HEADERS = ["Name", "Path", "Dimensions", "Shape", "Kind",
               "Element Type", "Type", "Unit", "Missing data", "Chunking", "Summary",
               "File name", "Item class", "Is open", "Error"]

    (COL_NODE_NAME, COL_NODE_PATH, COL_DIMS, COL_SHAPE, COL_KIND,
     COL_ELEM_TYPE, COL_TYPE, COL_UNIT, COL_MISSING_DATA, COL_CHUNKING, COL_SUMMARY,
     COL_FILE_NAME, COL_RTI_TYPE, COL_IS_OPEN, COL_EXCEPTION) = range(len(HEADERS))

    COL_DECORATION = COL_NODE_NAME  # Column number that contains the icon. None for no icons


    def __init__(self, parent=None):
        """ Constructor
        """
        super(RepoTreeModel, self).__init__(parent=parent)
        self._invisibleRootTreeItem = BaseRti(nodeName='<invisible-root>', iconColor='#FFFFFF')
        self._invisibleRootTreeItem.model = self
        self._isEditable = False


    def itemData(self, treeItem, column, role=Qt.DisplayRole):
        """ Returns the data stored under the given role for the item. O
        """
        if role == Qt.DisplayRole:
            if column == self.COL_NODE_NAME:
                return treeItem.nodeName
            elif column == self.COL_NODE_PATH:
                return treeItem.nodePath
            elif column == self.COL_DIMS:
                if treeItem.isSliceable:
                    return " × ".join(str(elem) for elem in treeItem.dimensionNames)
                else:
                    return ""
            elif column == self.COL_SHAPE:
                if treeItem.isSliceable:
                    return " × ".join(str(elem) for elem in treeItem.arrayShape)
                else:
                    return ""
            elif column == self.COL_IS_OPEN:
                # Only show for RTIs that actually open resources.
                # TODO: this must be clearer. Use CanFetchChildren? Set is Open to None by default?
                if treeItem.hasChildren():
                    return str(treeItem.isOpen)
                else:
                    return ""
            elif column == self.COL_KIND:
                return treeItem.dimensionality
            elif column == self.COL_ELEM_TYPE:
                return treeItem.elementTypeName
            elif column == self.COL_TYPE:
                return treeItem.typeName
            elif column == self.COL_FILE_NAME:
                return treeItem.fileName if hasattr(treeItem, 'fileName') else ''
            elif column == self.COL_UNIT:
                return treeItem.unit
            elif column == self.COL_CHUNKING:
                return treeItem.chunksString
            elif column == self.COL_MISSING_DATA:
                return to_string(treeItem.missingDataValue, noneFormat='') # empty str for Nones
            elif column == self.COL_RTI_TYPE:
                return type_name(treeItem)
            elif column == self.COL_EXCEPTION:
                return str(treeItem.exception) if treeItem.exception else ''
            elif column == self.COL_SUMMARY:
                return treeItem.summary
            else:
                raise ValueError("Invalid column: {}".format(column))

        elif role == Qt.ToolTipRole:
            if treeItem.exception:
                return str(treeItem.exception)
            if column == self.COL_NODE_NAME:
                return treeItem.nodePath # Also path when hovering over the name
            elif column == self.COL_NODE_PATH:
                return treeItem.nodePath
            elif column == self.COL_DIMS:
                if treeItem.isSliceable:
                    if treeItem.dimensionPaths is not None:
                        return " ×\n".join(str(elem) for elem in treeItem.dimensionPaths)
                    else:
                        return " × ".join(str(elem) for elem in treeItem.dimensionNames)
                else:
                    return ""
            elif column == self.COL_SHAPE:
                if treeItem.isSliceable:
                    return " × ".join(str(elem) for elem in treeItem.arrayShape)
                else:
                    return ""
            elif column == self.COL_UNIT:
                return treeItem.unit
            elif column == self.COL_CHUNKING:
                return treeItem.chunksString
            elif column == self.COL_MISSING_DATA:
                return to_string(treeItem.missingDataValue, noneFormat='') # empty str for Nones
            elif column == self.COL_RTI_TYPE:
                return type_name(treeItem)
            elif column == self.COL_KIND:
                return treeItem.dimensionality
            elif column == self.COL_ELEM_TYPE:
                return treeItem.elementTypeName
            elif column == self.COL_TYPE:
                return treeItem.typeName
            elif column == self.COL_FILE_NAME:
                return treeItem.fileName if hasattr(treeItem, 'fileName') else ''
            elif column == self.COL_SUMMARY:
                return treeItem.summary
            else:
                return None

        elif role == Qt.TextAlignmentRole:
            if column == self.COL_SUMMARY:
                return ALIGN_RIGHT
            else:
                return ALIGN_LEFT

        else:
            return super(RepoTreeModel, self).itemData(treeItem, column, role=role)


    def canFetchMore(self, parentIndex):
        """ Returns true if there is more data available for parent; otherwise returns false.
        """
        parentItem = self.getItem(parentIndex)
        if not parentItem:
            return False

        return parentItem.canFetchChildren()


    def fetchMore(self, parentIndex):
        """ Fetches any available data for the items with the parent specified by the parent index.
        """
        parentItem = self.getItem(parentIndex)
        if not parentItem:
            return

        if not parentItem.canFetchChildren():
            return

        for childItem in parentItem.fetchChildren():
            self.insertItem(childItem, parentIndex=parentIndex)

        # Check that Rti implementation correctly sets canFetchChildren
        assert not parentItem.canFetchChildren(), \
            "not all children fetched: {}".format(parentItem)


    def findFileRtiIndex(self, childIndex):
        """ Traverses the tree upwards from the item at childIndex until the tree
            item is found that represents the file the item at childIndex
        """
        parentIndex = childIndex.parent()
        if not parentIndex.isValid():
            return childIndex
        else:
            parentItem = self.getItem(parentIndex)
            childItem = self.getItem(childIndex)
            if parentItem.fileName == childItem.fileName:
                return self.findFileRtiIndex(parentIndex)
            else:
                return childIndex


    def reloadFileAtIndex(self, itemIndex, rtiRegItem=None):
        """ Reloads the item at the index by removing the repo tree item and inserting a new one.

            The new item will have by of type rtiClass. If rtiRegItem is None (the default), the
            new rtiClass will be the same as the old one.
        """
        fileRtiParentIndex = itemIndex.parent()
        fileRti = self.getItem(itemIndex)
        position = fileRti.childNumber()
        fileName = fileRti.fileName

        # Delete old RTI and Insert a new one instead.
        self.deleteItemAtIndex(itemIndex) # this will close the items resources.

        if rtiRegItem is None:
            # Do NOT autodetect but use the class from the the RTI that's being replaced.
            rtiClass = type(fileRti)
            logger.debug("Recreating class from previous RTI class: {}".format(rtiClass))
            repoTreeItem = rtiClass.createFromFileName(fileName, fileRti.iconColor)

            assert repoTreeItem.parentItem is None, "repoTreeItem {!r}".format(repoTreeItem)
            return self.insertItem(repoTreeItem, position=position, parentIndex=fileRtiParentIndex)

        else:
            return self.loadFile(fileName, rtiRegItem, position=position,
                                 parentIndex=fileRtiParentIndex)


    def loadFile(self, fileName, rtiRegItem,
                 position=None, parentIndex=QtCore.QModelIndex()):
        """ Loads a file in the repository as a repo tree item of class rtiClass.
            Autodetects the RTI type if rtiClass is None.
            If position is None the child will be appended as the last child of the parent.
            Returns the index of the newly inserted RTI
        """
        check_class(rtiRegItem, RtiRegItem, allow_none=True)
        fileName = normRealPath(fileName)
        logger.info("Loading data from: {!r}".format(fileName))

        rtiClass = rtiRegItem.getClass(tryImport=True) if rtiRegItem else None
        if rtiClass is None:
            repoTreeItem = createRtiFromFileName(fileName)
        else:
            repoTreeItem = rtiClass.createFromFileName(fileName, rtiRegItem.iconColor)

        assert repoTreeItem.parentItem is None, "repoTreeItem {!r}".format(repoTreeItem)
        return self.insertItem(repoTreeItem, position=position, parentIndex=parentIndex)





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository Tree Items (RTIs) for Exdir data.

"""
from __future__ import print_function, absolute_import

import logging, os
import collections
import exdir
import numpy as np

try:
    from collections import MutableMapping
except:
    from collections.abc import MutableMapping # Python > 3.10

from argos.repo.iconfactory import RtiIconFactory, ICON_COLOR_UNDEF
from argos.repo.baserti import BaseRti, shapeToSummary
from argos.repo.filesytemrtis import createRtiFromFileName
from argos.repo.rtiplugins.hdf5 import dimNamesFromDataset  # We can reuse it, the exdir module follows the h5py API.
from argos.utils.cls import check_class, is_an_array
from argos.utils.defs import DIM_TEMPLATE, SUB_DIM_TEMPLATE
from argos.utils.masks import maskedEqual


logger = logging.getLogger(__name__)




def dataSetElementType(exdirDataset):
    """ Returns a string describing the element type of the dataset
    """
    dtype =  exdirDataset.dtype

    if dtype.names:
        return '<structured>'
    else:
        if dtype.metadata and 'vlen' in dtype.metadata:
            vlen_type = dtype.metadata['vlen']
            try:
                return "<vlen {}>".format(vlen_type.__name__)  # when vlen_type is a type
            except AttributeError: #
                return "<vlen {}>".format(vlen_type.name)      # when vlen_type is a dtype

    return str(dtype)


def dataSetUnit(exdirDataset):
    """ Returns the unit of the exdirDataset by looking in the attributes.

        It searches in the attributes for one of the following keys:
        'unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'. If these are not found, the empty
        string is returned.

        Always returns a string
    """
    attributes = exdirDataset.attrs.to_dict()
    if not attributes:
        return '' # a premature optimization :-)

    for key in ('unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'):
        if key in attributes:
            return attributes[key]
    # Not found
    return ''


def dataSetMissingValue(exdirDataset):
    """ Returns the missingData given a Exdir dataset

        Looks for one of the following attributes: _FillValue, missing_value, MissingValue,
        missingValue. Returns None if these attributes are not found.
    """
    attributes = exdirDataset.attrs.to_dict()
    if not attributes:
        return None # a premature optimization :-)

    for key in ('missing_value', 'MissingValue', 'missingValue', 'FillValue', '_FillValue'):
        if key in attributes:
            missingDataValue = attributes[key]
            if is_an_array(missingDataValue) and len(missingDataValue) == 1:
                return missingDataValue[0] # In case of HDF-EOS and NetCDF files
            else:
                return missingDataValue
    return None


def flattenDict(d, parent_key='', sep='/'):
    """ Returns a flatten dictionary given a nested dictionary.

        The nested keys are be separated by sep.
    """
    items = []
    if isinstance(d,list): d = {str(i) : v for i, v in enumerate(d)}
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flattenDict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



class ExdirScalarRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a scalar HDF-5 variable.

    """
    _defaultIconGlyph = RtiIconFactory.SCALAR

    def __init__(self, exdirDataset, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(ExdirScalarRti, self).__init__(
            nodeName=nodeName, fileName=fileName, iconColor=iconColor)
        check_class(exdirDataset, exdir.Dataset)
        self._exdirDataset = exdirDataset


    def hasChildren(self):
        """ Returns False. Leaf nodes never have children. """
        return False


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        array = np.array([self._exdirDataset[()]]) # slice with empty tuple
        maskedArray = maskedEqual(array, self.missingDataValue)

        assert maskedArray.shape == (1, ), "Scalar wrapper shape mismatch: {}".format(array.shape)
        return maskedArray[index] # Use the index to ensure the slice has the correct shape


    @property
    def arrayShape(self):
        """ Returns the shape of the wrapper array. Will always be the tuple (1, )
        """
        return (1, )


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return "scalar"


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetElementType(self._exdirDataset)


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return flattenDict(self._exdirDataset.attrs.to_dict()) # add to_dict() ?


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        return dataSetUnit(self._exdirDataset)


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        return dataSetMissingValue(self._exdirDataset)


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI. In this case the scalar as a string
        """
        return str(self._exdirDataset[()])



class ExdirFieldRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a field in a structured HDF-5 variable.
    """
    _defaultIconGlyph = RtiIconFactory.FIELD

    def __init__(self, exdirDataset, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor.
            The name of the field must be given to the nodeName parameter.
        """
        super(ExdirFieldRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(exdirDataset, exdir.Dataset)
        self._exdirDataset = exdirDataset


    def hasChildren(self):
        """ Returns False. Field nodes never have children.
        """
        return False


    @property
    def attributes(self):
        """ The attributes dictionary.
            Returns the attributes of the variable that contains this field.
        """
        return flattenDict(self._exdirDataset.attrs.to_dict()) # add to_dict() ?


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Applies the index on the NCDF variable that contain this field and then selects the
            current field. In pseudo-code, it returns: self.ncVar[index][self.nodeName].

            If the field itself contains a sub-array it returns:
                self.ncVar[mainArrayIndex][self.nodeName][subArrayIndex]
        """
        mainArrayNumDims = len(self._exdirDataset.shape)
        mainIndex = index[:mainArrayNumDims]
        mainArray = self._exdirDataset.__getitem__(mainIndex)
        fieldArray = mainArray[self.nodeName]
        subIndex = tuple([Ellipsis]) + index[mainArrayNumDims:]
        slicedArray = fieldArray[subIndex]

        return maskedEqual(slicedArray, self.missingDataValue)


    @property
    def nDims(self):
        """ The number of dimensions of the underlying array
        """
        return len(self.arrayShape) # h5py datasets don't have an ndim property


    @property
    def _subArrayShape(self):
        """ Returns the shape of the sub-array
            An empty tuple is returned for regular fields, which have no sub array.
        """
        if self._exdirDataset.dtype.fields is None:
            return tuple() # regular field
        else:
            fieldName = self.nodeName
            fieldDtype = self._exdirDataset.dtype.fields[fieldName][0]
            return fieldDtype.shape


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
            If the field contains a subarray the shape may be longer than 1.
        """
        return self._exdirDataset.shape + self._subArrayShape


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return "field"


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        fieldName = self.nodeName
        return str(self._exdirDataset.dtype.fields[fieldName][0])


    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying NCDF variable
        """
        nSubDims = len(self._subArrayShape)
        subArrayDims = [SUB_DIM_TEMPLATE.format(dimNr) for dimNr in range(nSubDims)]
        return dimNamesFromDataset(self._exdirDataset) + subArrayDims


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        unit = dataSetUnit(self._exdirDataset)
        fieldNames = self._exdirDataset.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(unit, '__len__') and len(unit) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return unit[idx]
        else:
            return unit


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        value = dataSetMissingValue(self._exdirDataset)
        fieldNames = self._exdirDataset.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(value, '__len__') and len(value) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return value[idx]
        else:
            return value


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI.  E.g. 'array 20 x 30' elements.
        """
        return shapeToSummary(self.arrayShape)



class ExdirDatasetRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a HDF5 dataset.

        This includes dimenions scales, which are then displayed with a different icon.
    """
    _defaultIconGlyph = RtiIconFactory.ARRAY # the iconGlyph property is overridden below

    def __init__(self, exdirDataset, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(ExdirDatasetRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(exdirDataset, exdir.Dataset)
        self._exdirDataset = exdirDataset
        self._isStructured = bool(self._exdirDataset.dtype.names)

        exdirDir = str(self._exdirDataset.directory) # convert to string for Python 3.5
        fileNames = os.listdir(os.path.abspath(exdirDir))
        absFileNames = [os.path.join(exdirDir, fn) for fn in fileNames]
        self._hasRaws = sum([os.path.isdir(f) for f in absFileNames]) > 0

    def hasChildren(self):
        """ Returns True if the variable has a structured type, otherwise returns False.
        """
        return self._isStructured or self._hasRaws


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Passes the index through to the underlying dataset.
            Converts to a masked array using the missing data value as fill_value
        """
        return maskedEqual(self._exdirDataset.__getitem__(index), self.missingDataValue)


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
        """
        return self._exdirDataset.shape


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return "array"


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetElementType(self._exdirDataset)


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return flattenDict(self._exdirDataset.attrs.to_dict()) #add .to_dict() ?


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        return dataSetUnit(self._exdirDataset)


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        return dataSetMissingValue(self._exdirDataset)


    def _fetchAllChildren(self):
        """ Fetches all fields that this variable contains.
            Only variables with a structured data type can have fields.
        """
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        # Add fields
        if self._isStructured:
            for fieldName in self._exdirDataset.dtype.names:
                childItems.append(ExdirFieldRti(self._exdirDataset, nodeName=fieldName,
                                                fileName=self.fileName, iconColor=self.iconColor))

        # Add raw directories
        if self._hasRaws:
            exdirDir = str(self._exdirDataset.directory) # convert to string for Python 3.5
            fileNames = os.listdir(os.path.abspath(exdirDir))
            absFileNames = [os.path.join(exdirDir, fn) for fn in fileNames]

            for fileName, absFileName in zip(fileNames, absFileNames):
                if os.path.isdir(absFileName) and not fileName.startswith('.'):
                    childItems.append(ExdirRawRti(
                        self._exdirDataset.require_raw(fileName), nodeName=fileName,
                        fileName=self.fileName, iconColor=self.iconColor))

        return childItems


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI.  E.g. 'array 20 x 30' elements.
        """
        return shapeToSummary(self.arrayShape)


class ExdirRawRti(BaseRti):
    """ Repository Tree Item (RTI) that contains an Exdir Raw.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER

    def __init__(self, exdirRaw, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(ExdirRawRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(exdirRaw, exdir.Raw, allow_none=True)

        self._exdirRaw = exdirRaw


    def _fetchAllChildren(self): # Raw is treated like a directory
        """ Fetches all sub groups and variables that this group contains.
        """
        assert self._exdirRaw is not None, "dataset undefined (file not opened?)"
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []
        exdirDir = str(self._exdirRaw.directory) # convert to string for Python 3.5
        absPaths = os.path.abspath(exdirDir)
        fileNames = os.listdir(absPaths)
        absFileNames = [os.path.join(exdirDir, fn) for fn in fileNames]

        for fileName, absFileName in zip(fileNames, absFileNames):
            if not fileName.startswith('.'):
                childItem = createRtiFromFileName(absFileName)
                childItems.append(childItem)

        return childItems



class ExdirGroupRti(BaseRti):
    """ Repository Tree Item (RTI) that contains an Exdir group.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER

    def __init__(self, exdirGroup, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(ExdirGroupRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(exdirGroup, exdir.Group, allow_none=True)

        self._exdirGroup = exdirGroup


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return flattenDict(self._exdirGroup.attrs.to_dict()) if self._exdirGroup else {}


    def _fetchAllChildren(self):
        """ Fetches all sub groups and variables that this group contains.
        """
        assert self._exdirGroup is not None, "dataset undefined (file not opened?)"
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        for childName, exdirChild in self._exdirGroup.items():
            if isinstance(exdirChild, exdir.Group):
                childItems.append(ExdirGroupRti(exdirChild, nodeName=childName,
                                                fileName=self.fileName, iconColor=self.iconColor))

            elif isinstance(exdirChild, exdir.Raw):
                childItems.append(ExdirRawRti(exdirChild, nodeName=childName,
                                              fileName=self.fileName, iconColor=self.iconColor))

            elif isinstance(exdirChild, exdir.Dataset):
                if len(exdirChild.shape) == 0:
                    childItems.append(ExdirScalarRti(
                        exdirChild, nodeName=childName, fileName=self.fileName,
                        iconColor=self.iconColor))
                else:
                    childItems.append(ExdirDatasetRti(
                        exdirChild, nodeName=childName, fileName=self.fileName,
                        iconColor=self.iconColor))
            else:
                logger.warning("Ignored {}. It has an unexpected Exdir type: {}"
                            .format(childName, type(exdir)))

        return childItems



class ExdirFileRti(ExdirGroupRti):
    """ Reads an Exdir file using the exdir package.
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(ExdirFileRti, self).__init__(None, nodeName, fileName=fileName, iconColor=iconColor)
        self._checkFileExists()

    def _openResources(self):
        """ Opens the root Dataset.
        """
        logger.info("Opening: {}".format(self._fileName))
        self._exdirGroup = exdir.File(self._fileName, mode='r')


    def _closeResources(self):
        """ Closes the root Dataset.
        """
        logger.info("Closing: {}".format(self._fileName))
        self._exdirGroup.close()
        self._exdirGroup = None



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Uses the Python Imaging Library (Pillow) to open an image
"""
from __future__ import absolute_import

import logging
import numpy as np

from PIL import Image

from argos.info import DEBUGGING
from argos.repo.iconfactory import RtiIconFactory, ICON_COLOR_UNDEF
from argos.repo.memoryrtis import ArrayRti, SliceRti

logger = logging.getLogger(__name__)

class PillowBandRti(SliceRti):
    """ Image band repo tree item. Will typically be a child of a PillowFileRti

        No dedicated constructor defined (reuses the ArrayRti constructor)
    """

    @property
    def dimensionNames(self):
        """ Returns ['Y', 'X'].
            The underlying array is expected to be 2-dimensional. If this is not the case we fall
            back on the default dimension names ['Dim-0', 'Dim-1', ...]
        """
        if self._array is None:
            return []

        if self._array.ndim != 2:
            # Defensive programming: fall back on default names
            msg = "Expected 2D image. Got: {}".format(self._array.ndim)
            if DEBUGGING:
                raise ValueError(msg)
            logger.warning(msg)
            return super(PillowBandRti, self).dimensionNames
        else:
            return ['Y', 'X']



class PillowFileRti(ArrayRti):
    """ Opens an image file with the Python Imaging Library (Pillow)

        See https://python-pillow.org/
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(PillowFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                            iconColor=iconColor)
        self._checkFileExists()
        self._bands = [] # image band names


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Reads the image and attributes from the underlying file
        """
        with Image.open(self._fileName) as image:
            self._array = np.asarray(image)
            self._bands = image.getbands()

            # Fill attributes. For now assume that the info item are not overridden by
            # the Image items.
            self._attributes = dict(image.info)
            self._attributes['Format'] = image.format
            self._attributes['Mode'] = image.mode
            self._attributes['Size'] = image.size
            self._attributes['Width'] = image.width
            self._attributes['Height'] = image.height


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None
        self._bands = []
        self._attributes = {}


    def _fetchAllChildren(self):
        """ Adds the bands as separate fields so they can be inspected easily.
        """
        bands = self._bands
        if len(bands) != self._array.shape[-1]:
            logger.warning("No bands added, bands != last_dim_lenght ({} !: {})"
                        .format(len(bands), self._array.shape[-1]))
            return []

        childItems = []
        for bandNr, band in enumerate(bands):
            bandItem = PillowBandRti(self._array[..., bandNr],
                                     nodeName=band, fileName=self.fileName,
                                     iconColor=self.iconColor, attributes=self._attributes)
            childItems.append(bandItem)
        return childItems


    @property
    def attributes(self):
        """ The attribute dictionary. Con
        """
        return self._attributes


    @property
    def dimensionNames(self):
        """ Returns ['Y', 'X', 'Band'].
            The underlying array is expected to be 3-dimensional. If this is not the case we fall
            back on the default dimension names ['Dim-0', 'Dim-1', ...]
        """
        if self._array is None:
            return []

        if self._array.ndim == 2:
            return ['Y', 'X']
        elif self._array.ndim == 3:
            return ['Y', 'X', 'Band']
        else:
            # Defensive programming: fall back on default names
            msg = "Expected 3D image. Got: {}".format(self._array.ndim)
            if DEBUGGING:
                raise ValueError(msg)
            logger.warning(msg)
            return super(PillowFileRti, self).dimensionNames




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Common functionality, look and feel for all table views in Argos.

    Control-C copies all data.
"""
from __future__ import print_function

import logging
from argos.qt import QtWidgets, Qt


logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901

class ArgosTableView(QtWidgets.QTableView):
    """ QTableView that defines common functionality for argos
    """
    def __init__(self, *args, **kwargs):
        """ Constructor
        """
        super(ArgosTableView, self).__init__(*args, **kwargs)


    def keyPressEvent(self, event):
        """ Overrides key press events to capture Ctrl-C
        """
        if event.key() == Qt.Key_C and event.modifiers() == Qt.ControlModifier:
            self.copySelectionToClipboard()
        else:
            super(ArgosTableView, self).keyPressEvent(event)
        return


    def copySelectionToClipboard(self):
        """ Copies selected cells to clipboard.

            Only works for ContiguousSelection
        """
        if not self.model():
            logger.warning("Table contains no data. Copy to clipboard aborted.")
            return

        if self.selectionMode() not in [QtWidgets.QTableView.SingleSelection,
                                        QtWidgets.QTableView.ContiguousSelection]:
            logger.warning("Copy to clipboard does not work for current selection mode: {}"
                           .format(self.selectionMode()))
            return

        selectedIndices = self.selectionModel().selectedIndexes()
        logger.info("Copying {} selected cells to clipboard.".format(len(selectedIndices)))

        # selectedIndexes() can return unsorted list so we sort it here to be sure.
        selectedIndices.sort(key=lambda idx: (idx.row(), idx.column()))

        # Unflatten indices into a list of list of indicides
        allIndices = []
        allLines = []
        lineIndices = []  # indices of current line
        prevRow = None
        for selIdx in selectedIndices:
            if prevRow != selIdx.row() and prevRow is not None: # new line
                allIndices.append(lineIndices)
                lineIndices = []
            lineIndices.append(selIdx)
            prevRow = selIdx.row()
        allIndices.append(lineIndices)
        del lineIndices

        # Convert to tab-separated lines so it can be pasted in Excel.
        lines = []
        for lineIndices in allIndices:
            line = '\t'.join([str(idx.data()) for idx in lineIndices])
            lines.append(line)
        txt = '\n'.join(lines)
        #print(txt)
        QtWidgets.QApplication.clipboard().setText(txt)


